@@ -2,4 +2,4 @@
  * String to embed in binaries to identify package
  */
 
-char pkg[] = "Package: netkit-ftp-0.10";
+char pkg[]="$NetKit: netkit-ftp-0.16 $";
